module com.fb;
import derelict.sdl.sdl;
import std.string;
import std.stdio;
import std.file;

const SDL_Color[] PALETTE = [
	{ 0,0,0 },       
	{ 63 << 2,63 << 2,63 << 2 },
	{ 26 << 2,13 << 2,10 << 2 },
	{ 28 << 2,41 << 2,44 << 2 },
	{ 27 << 2,15 << 2,33 << 2 },
	{ 22 << 2,35 << 2,16 << 2 },
	{ 13 << 2,10 << 2,30 << 2 },
	{ 46 << 2,49 << 2,27 << 2 },
	{ 27 << 2,19 << 2,9 << 2 },
	{ 16 << 2,14 << 2,0 << 2 },
	{ 38 << 2,25 << 2,22 << 2 },
	{ 17 << 2,17 << 2,17 << 2 },
	{ 27 << 2,27 << 2,27 << 2 },
	{ 38 << 2,52 << 2,33 << 2 },
	{ 27 << 2,23 << 2,45 << 2 },
	{ 37 << 2,37 << 2,37 << 2 } ];

enum Resolution { Res800x600, Res1024x768 }

class Video {
	private {
		SDL_Surface* surface;
		SDL_Overlay* overlay;
		SDL_Rect rect;
	}
	int height, width;
	float scalex, scaley;
	const int maxheight, maxwidth, arheight, arwidth;
	Resolution resolution;
	int fullscreen;
	bool useOverlay = false;
	bool keepAspect = false; 
	const bool realFullscreen = true;
	int sdlflags;
	this(Resolution res, int fs, bool p_useOverlay, bool p_keepAspect) {
		resolution = res;
		SDL_VideoInfo* vidinfo = SDL_GetVideoInfo();
		arheight = maxheight = vidinfo.current_h;
		arwidth = maxwidth = vidinfo.current_w;
		keepAspect = p_keepAspect;
		if(keepAspect) {
			arheight = cast(int)(maxwidth * 600.0/800.0);
			arwidth = maxwidth;
		}
		width = 800; height = 600;
		useOverlay = p_useOverlay;
		fullscreen = fs;
		enableFullscreen(fs > 0 ? true : false);
	}

	~this() {
		if(overlay !is null)
			SDL_FreeYUVOverlay(overlay);
		if(surface !is null)
			SDL_FreeSurface(surface);
	}
	
	void enableFullscreen(bool fs) {
		resize(fs);
		fullscreen = ((fs && !useOverlay) || (fs & realFullscreen)) ? SDL_FULLSCREEN : 0;
		sdlflags = SDL_SWSURFACE;
		//if(fs && useOverlay) sdlflags |= SDL_NOFRAME;
		sdlflags |= fullscreen;

		surface = SDL_SetVideoMode(width, height, 0, sdlflags); 
		if(surface is null) {
			throw new DisplayError("Unable to initialize graphics mode.");
		}
		SDL_SetPalette(surface, SDL_PHYSPAL|SDL_LOGPAL, 
					   cast(SDL_Color *)PALETTE, 0, 16);
		if(useOverlay) {
			makeOverlay([800, 600], [width, height]);
		}
		screen.refresh();
	}

	private void resize(bool maxres) {
		if(useOverlay && maxres) {
			if(keepAspect) {
				width = arwidth;
				height = arheight;
			}
			else {
				width = maxwidth;
				height = maxheight;
			}
		}
		else 
		{
			width = 800;
			height = 600;
		}
		scalex = 800.0 / width;
		scaley = 600.0 / height;
	}
	
	private void makeOverlay(int[] ovlRes, int[] scrRes) {
		if(overlay !is null)
			SDL_FreeYUVOverlay(overlay);
		overlay = SDL_CreateYUVOverlay(ovlRes[0], ovlRes[1], SDL_YV12_OVERLAY, surface);
		if (overlay is null) {
			throw new DisplayError("Couldn't initialize YUV overlay.");
		}
		rect.w = scrRes[0];
		rect.h = scrRes[1];
		rect.x = rect.y = 0;
	}
}

class Screen {
	const Uint16[] data, olddata;
	const int width, height;
	alias width w;
	alias height h;
	this(int xchars, int ychars) {
		width = xchars;
		height = ychars;
		data.length = xchars * ychars;
		olddata.length = xchars * ychars;
		refresh();
	}
	
	Uint16 getChar(int x, int y) {
		mixin(CHECKS);
		return data[x + y * width];
	}

	void setChar(int x, int y, Uint16 c) {
		mixin(CHECKS);
		data[x + y * width] = c;
		dirty = true;
	}

	void setColor(int x, int y, int fg, int bg) {
		mixin(CHECKS);
		Uint16* s = &data[x + y * width];
		*s &= 0xff;
		*s |= (fg << 8) | (bg << 12);
		dirty = true;
	}
	
	void setbg(int x, int y, int bg) {
		Uint16* s = &data[x + y * width];
		*s &= 0xfff;
		*s |= (bg << 12);
		dirty = true;
	}

	void setfg(int x, int y, int fg) {
		Uint16* s = &data[x + y * width];
		*s &= 0xf0ff;
		*s |= (fg << 8);
		dirty = true;
	}

	void clrtoeol(int y, int bg) {
		clrtoeol(0, y, bg);
	}

	void clrtoeol(int x, int y, int bg) {
		mixin(CHECKY);
		Uint16* s = &data[x + y * width];
		Uint16 v = cast(Uint16)(0x20 | (bg << 12));
		while(x++ < width) *s++ = v;
		dirty = true;
	}

	void clrbgtoeol(int y, int bg) {
		mixin(CHECKY);
		Uint16* s = &data[y * width];
		Uint16 v = cast(Uint16)(0x20 | (bg << 12));
		int i;
		
		while(i++ < width) {
			*s &= 0xfff;
			*s++ |= v;
		}
		dirty = true;
	}

	void clrscr() {
		data[] = 0x20;
	}

	void refresh() {
		olddata[] = 255;
	}

	void setcoltoeol(int y, int fg, int bg) {
		mixin(CHECKY);
	
		Uint16* s = &data[y * width];
		Uint16 v = cast(Uint16)((fg << 8) | (bg << 12));
		int i;
		
		while(i++ < width) {
			*s &= 255;
			*s |= v;
			s++;		
		}
	}
}

class DisplayError : Error {
	this(string msg) {
		super(msg ~ " " ~ std.string.toString(SDL_GetError()));
	}
}

Video video;
Screen screen;
const FONT_X = 8, FONT_Y = 14;
const ubyte[] font;
int mode;
bool dirty = false;
int border = 1;
const int Y_OFFSET;

const CHECKX = "assert(x >= 0 && x < width);";
const CHECKY = "assert(y >= 0 && y < height);";
const CHECKS = "assert(x + y >= 0 && x + y < video.width*video.height);";

static this() {
	void[] arr;
	font.length = 256*16;
	// realign font data
	const rawfont = import("font.psf");
	for(int i=0;i<256;i++) {
		font[i*16..i*16+14] = cast(ubyte[])rawfont[i*FONT_Y+4..i*FONT_Y+4+FONT_Y];
	}
	Y_OFFSET = 0;
}

void enableKeyRepeat() {
	SDL_EnableKeyRepeat(200, 10);
}

void disableKeyRepeat() {
	SDL_EnableKeyRepeat(0, 0);

}

bool sdlVideoInited = false;
void init(bool fullscreen, int m, bool yuv, bool aspect, string title) {
	if( SDL_Init(SDL_INIT_VIDEO) < 0) {
		throw new DisplayError("Couldn't initialize framebuffer.");
	}

	int mx, my, fs, i;
	mode = m;
	if(fullscreen)
		fs |= SDL_FULLSCREEN;
	mx = 800; my = 600;
	if(m > 0) {
		mx = 1024; my = 768; 
		border = 4;
	}
	int width = mx / FONT_X;
	int height = my / FONT_Y;

	screen = new Screen(width, height);

	if(!sdlVideoInited) video = new Video(Resolution.Res800x600, fullscreen, yuv, aspect);

	SDL_EnableKeyRepeat(200, 10);
	SDL_EnableUNICODE(1);

	SDL_WM_SetCaption(title.toStringz(),title.toStringz());

	//video.resize(false);
	
}

void cprint(int x, int y, int fg, int bg, char[] txt) {
	mixin(CHECKS);
	bool skipbg, skipfg;
	if(bg < 0) { skipbg = true; bg = 0; }
	if(fg < 0) { skipfg = true; fg = 0; }
	Uint16[] s = screen.data[x + y * screen.width .. x + y * screen.width + txt.length];
	Uint16 col = cast(Uint16)((fg << 8) | (bg << 12));
	foreach(i, char c; txt) {
		if(skipbg)
			col = cast(Uint16)((fg << 8) | (s[i] & 0xf000));
		if(skipfg)
			col = cast(Uint16)((bg << 12) | (s[i] & 0x0f00));
	
		s[i] = cast(Uint16)(c | col);
	}
	dirty = true;
}

void fprint(int x, int y, char[] str) {
	mixin(CHECKS);
	assert(str.length < 256);
	int pos1, pos2;
	int col = 0, bg, fg;
	int spos = x + y * screen.width;
	Uint16[] outb = screen.data[spos .. spos + str.length];

	int getcol(char c, int prev) {
		if(c == '+')
			return prev;
		return "0123456789abcdef".find(c);
	}
	while(pos1 < str.length) {
		if(str[pos1] == '`') {
			bg = getcol(str[pos1+1],outb[pos2] >> 12);
			fg = getcol(str[pos1+2],(outb[pos2] >> 8) & 15);
			if(fg >= 0 && bg >= 0) {
				col = (bg << 12) | (fg << 8);
				pos1 += 3;
				continue;
			}
		}
		outb[pos2] = (str[pos1] & 255) | col;
		pos2++;
		pos1++;
	}
	dirty = true;
}

void update() {
	if(video.useOverlay)
		updateYUVFrame();
	else updateSDLFrame();
}

void updateSDLFrame() {
	int x, y;
	int a,b,c;
	Uint16* bptr = &screen.data[0];
	Uint16* cptr = &screen.olddata[0];
	Uint32* sptr = cast(Uint32 *)video.surface.pixels + Y_OFFSET * screen.width * 8;
	Uint32* sp;
	Uint8* bp;
	Uint8 ubg, ufg;
		
	if (!dirty) return;
	dirty = false;

	SDL_LockSurface(video.surface);
  
	for(y = 0;y < screen.height; y++) {
		for(x = 0; x < screen.width; x++) {
			if(*bptr != *cptr) {
				*cptr = *bptr;
				sp = sptr;
				a = *bptr & 255;
				bp = &font[a * 16];
				ufg = (*bptr >> 8) & 15;
				ubg = (*bptr >> 12);
				int fgcolor = PALETTE[ufg].b << video.surface.format.Bshift | 
					(PALETTE[ufg].g << video.surface.format.Gshift) |
					(PALETTE[ufg].r << video.surface.format.Rshift);
				int bgcolor = PALETTE[ubg].b << video.surface.format.Bshift | 
					(PALETTE[ubg].g << video.surface.format.Gshift) |
					(PALETTE[ubg].r << video.surface.format.Rshift);
				for(c = 4; c < 18; c++, bp++) {
					b = *bp;
					Uint32* framePtr = sp;
					if(b & 0x80) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x40) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x20) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x10) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x08) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x04) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x02) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x01) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					sp += video.width - 8;
				}
			}
			sptr += 8;
			bptr++;
			cptr++;
		}
		sptr += video.width*13;
	}

	SDL_UnlockSurface(video.surface);
	SDL_Flip(video.surface);
}

void updateYUVFrame() {
	int x, y;
	int a,b,c;
	static Uint32[32] pixbuf;
	Uint16* bptr = &screen.data[0];
	Uint16* cptr = &screen.olddata[0];
	Uint32* sp;
	Uint8* bp;
	Uint8 ubg, ufg;
		
	if (!dirty) return;
	dirty = false;

	SDL_LockYUVOverlay(video.overlay);
	
	// clear bottom stripe if necessary (mode = 800x600)
	if(video.resolution == Resolution.Res800x600) {
		pixbuf[] = 0;
		for(x = 0; x < screen.width; x++) {
			for(y = 0; y < 12; y++) {
				RGBBlock2YUV(pixbuf, x * 8, screen.height * 14 + y);
			}
		}
	}
	
	for(y = 0;y < screen.height; y++) {
		for(x = 0; x < screen.width; x++) {
			if(*bptr != *cptr) {
				*cptr = *bptr;
				a = *bptr & 255;
				bp = &font[a * 16];
				ufg = (*bptr >> 8) & 15;
				ubg = (*bptr >> 12);
				int fgcolor = PALETTE[ufg].b << video.surface.format.Bshift | 
					(PALETTE[ufg].g << video.surface.format.Gshift) |
					(PALETTE[ufg].r << video.surface.format.Rshift);
				int bgcolor = PALETTE[ubg].b << video.surface.format.Bshift | 
					(PALETTE[ubg].g << video.surface.format.Gshift) |
					(PALETTE[ubg].r << video.surface.format.Rshift);
				for(c = 4; c < 18; c++, bp++) {
					sp = &pixbuf[0];
					b = *bp;
					Uint32* framePtr = sp;
					if(b & 0x80) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x40) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x20) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x10) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x08) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x04) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x02) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					if(b & 0x01) *(sp++) = fgcolor;
					else *(sp++) = bgcolor;
					RGBBlock2YUV(pixbuf, x * 8, y * 14 + c - 4);
					sp += video.width - 8;
				}
			}
			bptr++;
			cptr++;
		}
	}
	SDL_UnlockYUVOverlay(video.overlay);
	SDL_DisplayYUVOverlay(video.overlay, &video.rect);
}

private	void RGBBlock2YUV(Uint32[] source, int x, int y) {
	void RGB_to_YUV(Uint8 *rgb, Uint8* yuv) {
		yuv[0] = cast(int)(0.299*rgb[0] + 0.587*rgb[1] + 0.114*rgb[2]);
		yuv[2] = cast(int)((rgb[2]-yuv[0])*0.565 + 128);
		yuv[1] = cast(int)((rgb[0]-yuv[0])*0.713 + 128);
	}

	static Uint8[3] yuv;
	static Uint8*[3] plane;
	plane[0] = video.overlay.pixels[0] + video.overlay.pitches[0] * y + x;
	plane[1] = video.overlay.pixels[1] + video.overlay.pitches[1] * y / 2 + x / 2;
	plane[2] = video.overlay.pixels[2] + video.overlay.pitches[2] * y / 2 + x / 2;
	for(int xc = 0; xc < 8 && xc < video.overlay.w; xc++) {
		RGB_to_YUV(cast(Uint8*)&source[xc], cast(ubyte*)&yuv);
		*(plane[0]++) = yuv[0];
		if(xc % 2 == 0 && y % 2 == 0) {
			*(plane[1]++) = yuv[2];
			*(plane[2]++) = yuv[1];
		}
	}
}

int paddedStringLength(string s, char padchar) {
	int i;
	for(i = s.length - 1; i >= 0; i--) {
		if(s[i] != padchar) return i+1;
	}
	return 0;
}
	
Uint16 readkey() {
	SDL_Event evt;
	bool loop = true;
	
	while(loop) {
		while(SDL_PollEvent(&evt)) {
			if(evt.type == SDL_QUIT) {
				SDL_Quit();
				return 0;
			}
			if(evt.type == SDL_KEYDOWN) {
				loop = false;
				break;
			}
		}
		SDL_Delay(50);
	}
	return evt.key.keysym.unicode;
}

void hexdump(ubyte[] buf, int rowlen) {
	int c;
	foreach(b; buf) {
		writef("%02X ", b);
		c++;
		if(c >= rowlen) {
			c = 0;
			writefln();
		}
	}
	writefln();
}

string petString2D(ubyte* petstring) {
	string s;
	int i;
	s.length = 256;
	while(*petstring != '\0') {
		s[i] = *petstring;
		petstring++;
	}
	s.length = i;
	return s;
}